from django import forms


class Autocomplete(forms.Select):
    template_name = 'reactivated/autocomplete'

    def get_context(self, name, value, attrs):
        to_field_name = self.choices.field.to_field_name or 'pk'

        # context = forms.Widget.get_context(self, name, value, attrs)
        # self.choices.queryset = self.choices.queryset._clone()[:10]
        # context = super().get_context(name, value, attrs)
        context = forms.Widget.get_context(self, name, value, attrs)
        selected = self.choices.queryset.filter(**{to_field_name: value}).first() if value else None

        if selected is not None:
            context['widget']['selected'] = {"value": getattr(selected, to_field_name), "label": str(selected)}
        else:
            context['widget']['selected'] = None
        return context



from django.apps import AppConfig
from django.conf import settings

import logging
import importlib
import subprocess
import json
import os

from . import extract_views_from_urlpatterns

logger = logging.getLogger("django.server")


def get_types_schema() -> str:
    urlconf = importlib.import_module(settings.ROOT_URLCONF)
    urlpatterns = urlconf.urlpatterns

    from server.urls import urlpatterns
    from django_extensions.management.commands.show_urls import Command
    from django.contrib.admindocs.views import simplify_regex
    from django.urls import converters
    from django.urls.resolvers import RoutePattern
    import json

    converter_mapping = {
        converters.IntConverter: "number",
        converters.StringConverter: "string",
        converters.UUIDConverter: "string",
        converters.SlugConverter: "string",
        converters.PathConverter: "string",
    }

    urls = extract_views_from_urlpatterns(urlpatterns)
    reverse = {}

    for _, regex, name, pattern in urls:
        if not isinstance(pattern, RoutePattern):
            continue
        reverse[name or regex] = {
            "route": f"/{regex}",
            "args": {
                arg_name: converter_mapping.get(arg_converter.__class__, "string")
                for arg_name, arg_converter in pattern.converters.items()
            },
        }

    return json.dumps(reverse)


class ReactivatedConfig(AppConfig):
    name = "reactivated"

    def ready(self):
        """
        Django's dev server actually starts twice. So we prevent generation on
        the first start. TODO: handle noreload.
        """
        if settings.DEBUG is False:
            return

        is_server_started = "DJANGO_SEVER_STARTING" in os.environ

        if is_server_started is False:
            os.environ["DJANGO_SEVER_STARTING"] = "true"
            return

        logger.info("Generating interfaces and client side code")


        #: Note that we don't pass the file object to stdout, because otherwise
        # webpack gets confused with the half-written file when we make updates.
        # Maybe there's a way to force it to be a single atomic write? I tried
        # open('w+b', buffering=0) but no luck.
        process = subprocess.Popen(
            ["node", "./node_modules/reactivated/generator.js"],
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
        )
        out, error = process.communicate(get_types_schema().encode())

        with open("client/generated.tsx", "w+b") as output:
            output.write(out)
        logger.info("Finished generating.")



from django.conf import settings
from django.middleware.csrf import get_token
from django.template import TemplateDoesNotExist, TemplateSyntaxError
from django.template.backends.base import BaseEngine
from django.template.context import make_context
from django.utils.functional import cached_property
from django.utils.module_loading import import_string


import os
import simplejson
import markdown

from . import render_jsx_to_string


class JSX(BaseEngine):
    # Name of the subdirectory containing the templates for this engine
    # inside an installed application.
    # app_dirname = 'foobar'

    def __init__(self, params):
        # params = params.copy()
        options = params.pop('OPTIONS').copy()

        self.context_processors = options.pop('context_processors', [])
        super().__init__(params)

        # self.engine = foobar.Engine(**options)

    def from_string(self, template_code):
        return 'ABC'
        """
        try:
          return Template(self.engine.from_string(template_code))
        except foobar.TemplateCompilationFailed as exc:
            raise TemplateSyntaxError(exc.args)
        """

    def get_template(self, template_name):
        if template_name.endswith('.html'):
            jsx_template_name = template_name.replace('.html', '.tsx')

            if os.path.isfile(os.path.join(settings.BASE_DIR, 'client/templates', jsx_template_name)):
                return Template(jsx_template_name, self)

        if template_name.endswith('.tsx') or template_name.endswith('.jsx'):
            if os.path.isfile(os.path.join(settings.BASE_DIR, 'client/templates', template_name)):
                return Template(template_name, self)

        raise TemplateDoesNotExist([], backend=self)

        try:
            return Template(self.engine.get_template(template_name))
        except foobar.TemplateNotFound as exc:
            raise TemplateDoesNotExist(exc.args, backend=self)
        except foobar.TemplateCompilationFailed as exc:
            raise TemplateSyntaxError(exc.args)

    @cached_property
    def template_context_processors(self):
        return [import_string(path) for path in self.context_processors]


class Template:
    def __init__(self, jsx_template_name, backend):
        self.jsx_template_name = jsx_template_name
        self.backend = backend

    def get_props(self, context=None, request=None):
        if self.jsx_template_name == 'registration/login.tsx':
            return {
                'form': context['form'],
            }
        elif self.jsx_template_name == '404.tsx':
            return {
                'request_path': context['request_path'],
            }
        elif self.jsx_template_name == 'flatpages/default.tsx':
            flatpage = context['flatpage']

            return {
                'flatpage': {
                    'title': flatpage.title,
                    'url': flatpage.url,
                    'content': markdown.markdown(flatpage.content),
                },
            }
        else:
            return context


    def render(self, context=None, request=None):
        react_context = make_context({}, request, autoescape=False).__dict__
        props = context
        # request._is_jsx_response = True
        # props = self.get_props(context, request)
        template_name = self.jsx_template_name.replace('.tsx', '')

        return render_jsx_to_string(request, template_name, react_context, props)

    def render(self, context=None, request=None):
        template_name = self.jsx_template_name.replace('.tsx', '').replace('.jsx', '')

        from django.template.backends.utils import csrf_input_lazy, csrf_token_lazy

        props = context or {}
        context = {}

        if request is not None:
            context['request'] = request
            # context['csrf_input'] = csrf_input_lazy(request)
            context['csrf_token'] = str(csrf_token_lazy(request))

            for context_processor in self.backend.template_context_processors:
                context.update(context_processor(request))

            return render_jsx_to_string(request, template_name, context, props)

        assert False, "At this time, only templates with the request object can be rendered with reactivated"



from .widgets import Autocomplete

from django import forms as django_forms
from django.http import JsonResponse


def autocomplete(view_func):
    def wrapped_view(request, *args, **kwargs):
        response = view_func(request, *args, **kwargs)
        autocomplete = request.GET.get("autocomplete", None)
        query = request.GET.get("query", "")

        if autocomplete is None:
            return response

        for key, item in response.context_data.items():
            if isinstance(item, django_forms.BaseForm):
                form = item
                break
        else:
            form = None

        if (
            form is not None
            and autocomplete in form.fields
            and isinstance(form.fields[autocomplete], django_forms.ModelChoiceField)
            and isinstance(form.fields[autocomplete].widget, Autocomplete)
        ):
            results = [
                {"value": result.pk, "label": str(result)}
                for result in form.fields[autocomplete].queryset.filter(
                    name__icontains=query
                )[:50]
            ]

            return JsonResponse({"results": results})

        return response

    return wrapped_view
